library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.cpu_package.all;

entity cpu is
	port	(
		clk : in std_logic;
		rst : in std_logic;

		-- instruction memory
		-- read port
		mem_w_address_i : in std_logic_vector(IADDR_BUS - 1 downto 0);
		mem_w_data_i : in std_logic_vector(IDATA_BUS*2 - 1 downto 0);
		mem_w_enable_i : in std_logic;
		mem_w_out_o : out std_logic_vector(IDATA_BUS*2 - 1 downto 0);

		-- configuration register
		conf_reg : in std_logic_vector(CONF_BUS - 1 downto 0);

		-- cpu signals
		cpu_out : out std_logic_vector(DDATA_BUS - 1 downto 0);
		cpu_stop : out std_logic
		);
end cpu;

architecture cpu_arch of cpu is
signal fetch_w_enable : std_logic;
signal fetch_w_addr : std_logic_vector(IADDR_BUS downto 0);
signal fetch_r_address : std_logic_vector(IADDR_BUS downto 0); -- + 1 bit
signal fetch_r_out : std_logic_vector(IDATA_BUS - 1 downto 0);
-- out of fetch stage
signal fetch_out : std_logic_vector(IDATA_BUS - 1 downto 0);
signal ex_opc_s : std_logic_vector(OPC_BUS - 1 downto 0);
signal wb_opc_s : std_logic_vector(OPC_BUS - 1 downto 0);
signal id_alu_a_fw_s : std_logic;
signal id_push_s : std_logic;
signal id_stop_s : std_logic;
signal id_write_out_s : std_logic;
signal id_opc_s : std_logic_vector(OPC_BUS - 1 downto 0);
signal id_reg_0_s : std_logic_vector(DDATA_BUS - 1 downto 0);
signal id_reg_1_s : std_logic_vector(DDATA_BUS - 1 downto 0);
signal id_alu_func_s : std_logic_vector(FUNC_BUS - 1 downto 0);

signal ex_alu_out_s : std_logic_vector(DDATA_BUS - 1 downto 0);
signal ex_r_out_s : std_logic_vector(DDATA_BUS - 1 downto 0);
signal ex_w_out_s : std_logic_vector(DDATA_BUS - 1 downto 0);
signal ex_push_s : std_logic;
signal ex_stop_s : std_logic;
signal wb_out_s : std_logic_vector(DDATA_BUS - 1 downto 0);
signal mem_w_out_s : std_logic_vector(mem_w_out_o'range);
signal cpu_stop_s : std_logic;
signal cpu_out_s : std_logic_vector(DDATA_BUS - 1 downto 0);
begin
u_fetch : fetch
	port map (w_enable => fetch_w_enable,
		w_addr => fetch_w_addr,
		r_address => fetch_r_address,
		r_out => fetch_r_out,
		fetch_out => fetch_out,
		clk => clk,
		rst => rst
		);
u_imem : mem
	port map (r_address => fetch_r_address,
		r_out => fetch_r_out,
		w_address => mem_w_address_i,
		w_data => mem_w_data_i,
		w_enable => mem_w_enable_i,
		w_out => mem_w_out_s,
		clk => clk
		);
mem_w_out_o <= mem_w_out_s;

u_decode : decode
	port map (
		clk => clk,
		rst => rst,
		instr_i => fetch_out,
		wb_out_i => wb_out_s,
		ex_opc_i => id_opc_s,
		wb_opc_i => wb_opc_s,
		id_alu_a_fw_o => id_alu_a_fw_s,
		id_push_o => id_push_s,
		id_stop_o => id_stop_s,
		id_write_out_o => id_write_out_s,
		id_opc_o => id_opc_s,
		id_reg_0_o => id_reg_0_s,
		id_reg_1_o => id_reg_1_s,
		id_alu_func_o => id_alu_func_s
		);
u_execute : execute
	port map (
		clk => clk,
		rst => rst,
		conf_reg => conf_reg,
		id_reg_0 => id_reg_0_s,
		id_reg_1 => id_reg_1_s,
		id_alu_func => id_alu_func_s,
		id_alu_a_fw => id_alu_a_fw_s,
		id_opc => id_opc_s,
		id_push => id_push_s,
		id_stop => id_stop_s,
		id_write_out => id_write_out_s,
		ex_alu_out => ex_alu_out_s,
		ex_r_out => ex_r_out_s(0),
		ex_w_out => ex_w_out_s(0),
		ex_push => ex_push_s,
		ex_stop => ex_stop_s,
		ex_opc => ex_opc_s,
		wb_out_i => wb_out_s
		);
u_wb : writeback
	port map (
		clk => clk,
		rst => rst,
		ex_alu_out_i => ex_alu_out_s,
		ex_r_out_i => ex_r_out_s,
		ex_w_out_i => ex_w_out_s,
		ex_push_i => ex_push_s,
		ex_stop_i => ex_stop_s,
		ex_opc_i => ex_opc_s,
		wb_opc_o => wb_opc_s,
		wb_out_o => wb_out_s,
		stop_o => cpu_stop_s,
		w_out_o => cpu_out_s
		);
cpu_stop <= cpu_stop_s;
cpu_out <= cpu_out_s;
end cpu_arch;