library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package cpu_package is
	constant REGFILE_SIZE : integer := 16;
	constant CONF_BUS : integer := 128;
	-- bus widths
	constant IADDR_BUS : integer := 4; -- trying with 16 double instructions
	constant IDATA_BUS : integer := 8;
	constant DDATA_BUS : integer := 1;
	constant STATUS_BUS : integer := 4;

	-- instruction format
	constant IFMT_OPC : std_logic_vector(7 downto 6) := (others => 'X');
	constant IFMT_INADDR : std_logic_vector(5 downto 0) := (others => 'X');
	constant IFMT_FUNC : std_logic_vector(5 downto 0) := (others => 'X');
	
	-- instruction field widths
	constant OPC_BUS : integer := (IFMT_OPC'left - IFMT_OPC'right) + 1;
	constant INADDR_BUS : integer := (IFMT_INADDR'left - IFMT_INADDR'right) + 1;
	constant FUNC_BUS : integer := (IFMT_FUNC'left - IFMT_FUNC'right) + 1;

	-- stack functions
	constant STACK_NOP : std_logic_vector := "00";
	constant STACK_PUSH : std_logic_vector := "01";
	constant STACK_POP1 : std_logic_vector := "10";
	constant STACK_POP2 : std_logic_vector := "11";

	-- opcodes
	constant OPC_IN : std_logic_vector := "10";
	constant OPC_OUT : std_logic_vector := "01";
	constant OPC_ALU : std_logic_vector := "00";

	-- alu functions, bit 3 is binary op selector
	constant BINARY_OP : integer := 3;
	constant ALU_FUNC_NOP : std_logic_vector := "000000";
	constant ALU_FUNC_NOT : std_logic_vector := "000001";
	constant ALU_FUNC_AND : std_logic_vector := "001001";
	constant ALU_FUNC_OR : std_logic_vector := "001010";
	constant ALU_FUNC_XOR : std_logic_vector := "001011";
	
	-- core types and constants
	constant N_CORES : integer := 32;
	type rst_a is array(natural range<>) of std_logic;
	type mem_w_address_a is array(natural range <>) of std_logic_vector(IADDR_BUS - 1 downto 0);
	type mem_w_data_a is array(natural range <>) of std_logic_vector(IDATA_BUS*2 - 1 downto 0);
	type mem_w_enable_a is array(natural range <>) of std_logic;
	type mem_w_out_a is array(natural range <>) of std_logic_vector(IDATA_BUS*2 - 1 downto 0);
	type cpu_out_a is array(natural range <>) of std_logic_vector(DDATA_BUS - 1 downto 0);
	type cpu_stop_a is array(natural range <>) of std_logic;

	-- core instruction memory address space
	constant MEM_CORE0 : std_logic_vector := B"0100_0000_0000";
	constant MEM_CORE1 : std_logic_vector := B"0100_0001_0000";
	constant MEM_CORE2 : std_logic_vector := B"0100_0010_0000";
	constant MEM_CORE3 : std_logic_vector := B"0100_0011_0000";
	constant MEM_CORE4 : std_logic_vector := B"0100_0100_0000";
	constant MEM_CORE5 : std_logic_vector := B"0100_0101_0000";
	constant MEM_CORE6 : std_logic_vector := B"0100_0110_0000";
	constant MEM_CORE7 : std_logic_vector := B"0100_0111_0000";
	constant MEM_CORE8 : std_logic_vector := B"0100_1000_0000";
	constant MEM_CORE9 : std_logic_vector := B"0100_1001_0000";
	constant MEM_CORE10 : std_logic_vector := B"0100_1010_0000";
	constant MEM_CORE11 : std_logic_vector := B"0100_1011_0000";
	constant MEM_CORE12 : std_logic_vector := B"0100_1100_0000";
	constant MEM_CORE13 : std_logic_vector := B"0100_1101_0000";
	constant MEM_CORE14 : std_logic_vector := B"0100_1110_0000";
	constant MEM_CORE15 : std_logic_vector := B"0100_1111_0000";
	constant MEM_CORE16 : std_logic_vector := B"0101_0000_0000";
	constant MEM_CORE17 : std_logic_vector := B"0101_0001_0000";
	constant MEM_CORE18 : std_logic_vector := B"0101_0010_0000";
	constant MEM_CORE19 : std_logic_vector := B"0101_0011_0000";
	constant MEM_CORE20 : std_logic_vector := B"0101_0100_0000";
	constant MEM_CORE21 : std_logic_vector := B"0101_0101_0000";
	constant MEM_CORE22 : std_logic_vector := B"0101_0110_0000";
	constant MEM_CORE23 : std_logic_vector := B"0101_0111_0000";
	constant MEM_CORE24 : std_logic_vector := B"0101_1000_0000";
	constant MEM_CORE25 : std_logic_vector := B"0101_1001_0000";
	constant MEM_CORE26 : std_logic_vector := B"0101_1010_0000";
	constant MEM_CORE27 : std_logic_vector := B"0101_1011_0000";
	constant MEM_CORE28 : std_logic_vector := B"0101_1100_0000";
	constant MEM_CORE29 : std_logic_vector := B"0101_1101_0000";
	constant MEM_CORE30 : std_logic_vector := B"0101_1110_0000";
	constant MEM_CORE31 : std_logic_vector := B"0101_1111_0000";

	
	-- alu
component alu is
	port	(
			alu_in_a : in std_logic_vector(DDATA_BUS - 1 downto 0);
			alu_in_b : in std_logic_vector(DDATA_BUS - 1 downto 0);
			alu_func : in std_logic_vector(FUNC_BUS - 1 downto 0);

			alu_out : out std_logic_vector(DDATA_BUS - 1 downto 0);
			alu_status : out std_logic_vector(STATUS_BUS - 1 downto 0)
		);
end component;
component decode is
	port	(
		clk : in std_logic;
		rst : in std_logic;
		instr_i : in std_logic_vector(IDATA_BUS - 1 downto 0);
		-- data from writeback
		wb_out_i : in std_logic_vector(DDATA_BUS - 1 downto 0);
		-- control signals
		-- input
		ex_opc_i : in std_logic_vector(OPC_BUS - 1 downto 0);
		wb_opc_i : in std_logic_vector(OPC_BUS - 1 downto 0);
		-- output
		id_alu_a_fw_o : out std_logic;
		id_push_o : out std_logic;
		id_stop_o : out std_logic;
		id_write_out_o : out std_logic;
		id_opc_o : out std_logic_vector(OPC_BUS - 1 downto 0);

		id_reg_0_o : out std_logic_vector(DDATA_BUS - 1 downto 0);
		id_reg_1_o : out std_logic_vector(DDATA_BUS - 1 downto 0);
		id_alu_func_o : out std_logic_vector(FUNC_BUS - 1 downto 0)
		);
end component;
component execute is
	port	(
		clk : in std_logic;
		rst : in std_logic;
		conf_reg : in std_logic_vector(CONF_BUS - 1 downto 0);
		-- instruction signals
		id_reg_0 : in std_logic_vector(DDATA_BUS - 1 downto 0);
		id_reg_1 : in std_logic_vector(DDATA_BUS - 1 downto 0);
		id_alu_func : in std_logic_vector(FUNC_BUS - 1 downto 0);
		-- controll signals
		id_alu_a_fw : in std_logic;
		id_opc : in std_logic_vector(OPC_BUS - 1 downto 0);
		id_push : in std_logic;
		id_stop : in std_logic;
		id_write_out : in std_logic;

		-- execute output registers
		ex_alu_out : out std_logic_vector(DDATA_BUS - 1 downto 0);
		ex_r_out : out std_logic;
		ex_w_out : out std_logic;
		ex_push : out std_logic;
		ex_stop : out std_logic;
		ex_opc : out std_logic_vector(OPC_BUS - 1 downto 0);

		-- from writeback
		wb_out_i : in std_logic_vector(DDATA_BUS - 1 downto 0)
		);
end component;
component fetch is
	port	(
		clk : in std_logic;
		rst : in std_logic;
		w_enable : in std_logic;
		w_addr : in std_logic_vector(IADDR_BUS downto 0);

		-- instruction memory
		-- read port
		r_address : out std_logic_vector(IADDR_BUS downto 0);
		r_out : in std_logic_vector(IDATA_BUS - 1 downto 0);
		fetch_out : out std_logic_vector(IDATA_BUS - 1 downto 0)
		);
end component;
component writeback is
	port	(
		clk : in std_logic;
		rst : in std_logic;
		-- registered stuff from execute
		ex_alu_out_i : in std_logic_vector(DDATA_BUS - 1 downto 0);
		ex_r_out_i : in std_logic_vector(DDATA_BUS - 1 downto 0);
		ex_w_out_i : in std_logic_vector(DDATA_BUS - 1 downto 0);
		ex_push_i : in std_logic;
		ex_stop_i : in std_logic;
		ex_opc_i : in std_logic_vector(OPC_BUS - 1 downto 0);

		-- control signals to decode
		wb_opc_o : out std_logic_vector(OPC_BUS - 1 downto 0);

		-- to decode and executek
		wb_out_o : out std_logic_vector(DDATA_BUS - 1 downto 0);
		-- out of cpu
		stop_o : out std_logic;
		w_out_o : out std_logic_vector(DDATA_BUS - 1 downto 0) 
		);
end component;
component io_reg is
	port (
		clk : in std_logic;
		rst : in std_logic;
		conf_reg : in std_logic_vector(CONF_BUS - 1 downto 0);
		conf_addr : in std_logic_vector(INADDR_BUS - 1 downto 0);
		w_enable : in std_logic;
		data_in : in std_logic_vector(DDATA_BUS - 1 downto 0);
		w_out : out std_logic;
		r_out : out std_logic);
end component;
component mem is
	generic	(
		MEM_ADDR_BUS : integer := IADDR_BUS;
		MEM_DATA_BUS : integer := IDATA_BUS * 2
		);
	port	(
		-- read port
		r_address : in std_logic_vector(MEM_ADDR_BUS downto 0); -- + 1
		r_out : out std_logic_vector(MEM_DATA_BUS/2 - 1 downto 0);

		-- write/read port
		w_address : in std_logic_vector(MEM_ADDR_BUS - 1 downto 0);
		w_data : in std_logic_vector(MEM_DATA_BUS - 1 downto 0);
		w_enable : in std_logic;
		w_out : out std_logic_vector(MEM_DATA_BUS - 1 downto 0);

		clk : in std_logic
		);
end component;
component cpu is
	port	(
		clk : in std_logic;
		rst : in std_logic;

		-- instruction memory
		-- read port
		mem_w_address_i : in std_logic_vector(IADDR_BUS - 1 downto 0);
		mem_w_data_i : in std_logic_vector(IDATA_BUS*2 - 1 downto 0);
		mem_w_enable_i : in std_logic;
		mem_w_out_o : out std_logic_vector(IDATA_BUS*2 - 1 downto 0);

		-- configuration register
		conf_reg : in std_logic_vector(CONF_BUS - 1 downto 0);

		-- cpu signals
		cpu_out : out std_logic_vector(DDATA_BUS - 1 downto 0);
		cpu_stop : out std_logic
		);
end component;
component ca_cpu is
	port (
			clk : in std_logic;
			rst : in std_logic;
			-- hold reset high until you want to start the cores
			core_rst : in std_logic;
			-- memory
			mem_w_address_array : in mem_w_address_a(0 to N_CORES - 1);
			mem_w_data_array : in mem_w_data_a(0 to N_CORES - 1);
			mem_w_enable_array : in mem_w_enable_a(0 to N_CORES - 1);
			mem_w_out_array : out mem_w_out_a(0 to N_CORES - 1);
			-- old state, this should be registered
			conf_reg_i : in std_logic_vector(CONF_BUS - 1 downto 0);
			-- new state, registered output
			conf_reg_o : out std_logic_vector(CONF_BUS - 1 downto 0);
			-- done status flag, means that all cores are finished
			done_o : out std_logic
			--cpu_out_o : out cpu_out_a(0 to N_CORES - 1)
			);
end component;
end cpu_package;
